# Some simple queue module tests, plus some failure conditions
# to ensure the Queue locks remain stable.
import asyncio
import queue
import threading
import time
from unittest.mock import patch

import pytest

import janus

QUEUE_SIZE = 5


def qfull(q):
    return q._parent._maxsize > 0 and q.qsize() == q._parent._maxsize


# A thread to run a function that unclogs a blocked Queue.
class _TriggerThread(threading.Thread):
    def __init__(self, fn, args):
        self.fn = fn
        self.args = args
        self.startedEvent = threading.Event()
        threading.Thread.__init__(self)

    def run(self):
        # The sleep isn't necessary, but is intended to give the blocking
        # function in the main thread a chance at actually blocking before
        # we unclog it.  But if the sleep is longer than the timeout-based
        # tests wait in their blocking functions, those tests will fail.
        # So we give them much longer timeout values compared to the
        # sleep here (I aimed at 10 seconds for blocking functions --
        # they should never actually wait that long - they should make
        # progress as soon as we call self.fn()).
        time.sleep(0.1)
        self.startedEvent.set()
        self.fn(*self.args)


# Execute a function that blocks, and in a separate thread, a function that
# triggers the release.  Returns the result of the blocking function.  Caution:
# block_func must guarantee to block until trigger_func is called, and
# trigger_func must guarantee to change queue state so that block_func can make
# enough progress to return.  In particular, a block_func that just raises an
# exception regardless of whether trigger_func is called will lead to
# timing-dependent sporadic failures, and one of those went rarely seen but
# undiagnosed for years.  Now block_func must be unexceptional.  If block_func
# is supposed to raise an exception, call do_exceptional_blocking_test()
# instead.


class BlockingTestMixin:
    def do_blocking_test(self, block_func, block_args, trigger_func, trigger_args):
        self.t = _TriggerThread(trigger_func, trigger_args)
        self.t.start()
        self.result = block_func(*block_args)
        # If block_func returned before our thread made the call, we failed!
        if not self.t.startedEvent.is_set():
            pytest.fail("blocking function '%r' appeared not to block" % block_func)
        self.t.join(10)  # make sure the thread terminates
        if self.t.is_alive():
            pytest.fail("trigger function '%r' appeared to not return" % trigger_func)
        return self.result

    # Call this instead if block_func is supposed to raise an exception.
    def do_exceptional_blocking_test(
        self,
        block_func,
        block_args,
        trigger_func,
        trigger_args,
        expected_exception_class,
    ):
        self.t = _TriggerThread(trigger_func, trigger_args)
        self.t.start()
        try:
            try:
                block_func(*block_args)
            except expected_exception_class:
                raise
            else:
                pytest.fail("expected exception of kind %r" % expected_exception_class)
        finally:
            self.t.join(10)  # make sure the thread terminates
            if self.t.is_alive():
                pytest.fail(
                    "trigger function '%r' appeared to not return" % trigger_func
                )
            if not self.t.startedEvent.is_set():
                pytest.fail("trigger thread ended but event never set")


class BaseQueueTestMixin(BlockingTestMixin):
    cum = 0
    cumlock = threading.Lock()

    def simple_queue_test(self, _q):
        q = _q.sync_q
        if q.qsize():
            raise RuntimeError("Call this function with an empty queue")
        assert q.empty()
        assert not q.full()
        # I guess we better check things actually queue correctly a little :)
        q.put(111)
        q.put(333)
        q.put(222)
        target_order = dict(
            Queue=[111, 333, 222],
            LifoQueue=[222, 333, 111],
            PriorityQueue=[111, 222, 333],
        )
        actual_order = [q.get(), q.get(), q.get()]
        assert actual_order == target_order[_q.__class__.__name__]
        for i in range(QUEUE_SIZE - 1):
            q.put(i)
            assert q.qsize()
        assert not qfull(q)
        last = 2 * QUEUE_SIZE
        full = 3 * 2 * QUEUE_SIZE
        q.put(last)
        assert qfull(q)
        assert not q.empty()
        assert q.full()
        try:
            q.put(full, block=0)
            pytest.fail("Didn't appear to block with a full queue")
        except queue.Full:
            pass
        try:
            q.put(full, timeout=0.01)
            pytest.fail("Didn't appear to time-out with a full queue")
        except queue.Full:
            pass
        # Test a blocking put
        self.do_blocking_test(q.put, (full,), q.get, ())
        self.do_blocking_test(q.put, (full, True, 10), q.get, ())
        # Empty it
        for i in range(QUEUE_SIZE):
            q.get()
        assert not q.qsize()
        try:
            q.get(block=0)
            pytest.fail("Didn't appear to block with an empty queue")
        except queue.Empty:
            pass
        try:
            q.get(timeout=0.01)
            pytest.fail("Didn't appear to time-out with an empty queue")
        except queue.Empty:
            pass
        # Test a blocking get
        self.do_blocking_test(q.get, (), q.put, ("empty",))
        self.do_blocking_test(q.get, (True, 10), q.put, ("empty",))

    def worker(self, q):
        try:
            while True:
                x = q.get()
                if x < 0:
                    q.task_done()
                    return
                with self.cumlock:
                    self.cum += x
                q.task_done()
        except Exception as ex:
            from traceback import print_exc

            print_exc(ex)

    def queue_join_test(self, q):
        self.cum = 0
        for i in (0, 1):
            threading.Thread(target=self.worker, args=(q,)).start()
        for i in range(100):
            q.put(i)
        q.join()
        assert self.cum == sum(range(100))
        for i in (0, 1):
            q.put(-1)  # instruct the threads to close
        q.join()  # verify that you can join twice

    @pytest.mark.asyncio
    async def test_queue_task_done(self):
        # Test to make sure a queue task completed successfully.
        _q = self.type2test()
        q = _q.sync_q
        try:
            q.task_done()
        except ValueError:
            pass
        else:
            pytest.fail("Did not detect task count going negative")
        _q.close()
        await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_queue_join(self):
        # Test that a queue join()s successfully, and before anything else
        # (done twice for insurance).
        _q = self.type2test()
        q = _q.sync_q
        self.queue_join_test(q)
        self.queue_join_test(q)
        try:
            q.task_done()
        except ValueError:
            pass
        else:
            pytest.fail("Did not detect task count going negative")
        finally:
            _q.close()
            await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_simple_queue(self):
        # Do it a couple of times on the same queue.
        # Done twice to make sure works with same instance reused.
        _q = self.type2test(QUEUE_SIZE)
        self.simple_queue_test(_q)
        self.simple_queue_test(_q)
        _q.close()
        await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_negative_timeout_raises_exception(self):
        _q = self.type2test(QUEUE_SIZE)
        q = _q.sync_q
        with pytest.raises(ValueError):
            q.put(1, timeout=-1)
        with pytest.raises(ValueError):
            q.get(1, timeout=-1)
        _q.close()
        await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_nowait(self):
        _q = self.type2test(QUEUE_SIZE)
        q = _q.sync_q
        for i in range(QUEUE_SIZE):
            q.put_nowait(1)
        with pytest.raises(queue.Full):
            q.put_nowait(1)

        for i in range(QUEUE_SIZE):
            q.get_nowait()
        with pytest.raises(queue.Empty):
            q.get_nowait()
        _q.close()
        await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_shrinking_queue(self):
        # issue 10110
        _q = self.type2test(3)
        q = _q.sync_q
        q.put(1)
        q.put(2)
        q.put(3)
        with pytest.raises(queue.Full):
            q.put_nowait(4)
        assert q.qsize() == 3
        q._maxsize = 2  # shrink the queue
        with pytest.raises(queue.Full):
            q.put_nowait(4)
        _q.close()
        await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_maxsize(self):
        # Test to make sure a queue task completed successfully.
        _q = self.type2test(5)
        q = _q.sync_q
        assert q.maxsize == 5
        _q.close()
        await _q.wait_closed()


class TestQueue(BaseQueueTestMixin):
    type2test = janus.Queue


class TestLifoQueue(BaseQueueTestMixin):
    type2test = janus.LifoQueue


class TestPriorityQueue(BaseQueueTestMixin):
    type2test = janus.PriorityQueue


# A Queue subclass that can provoke failure at a moment's notice :)
class FailingQueueException(Exception):
    pass


class FailingQueue(janus.Queue):
    def __init__(self, *args, **kwargs):
        self.fail_next_put = False
        self.fail_next_get = False
        super().__init__(*args, **kwargs)

    def _put(self, item):
        if self.fail_next_put:
            self.fail_next_put = False
            raise FailingQueueException("You Lose")
        return super()._put(item)

    def _get(self):
        if self.fail_next_get:
            self.fail_next_get = False
            raise FailingQueueException("You Lose")
        return super()._get()


class TestFailingQueue(BlockingTestMixin):
    def failing_queue_test(self, _q):
        q = _q.sync_q
        if q.qsize():
            raise RuntimeError("Call this function with an empty queue")
        for i in range(QUEUE_SIZE - 1):
            q.put(i)
        # Test a failing non-blocking put.
        _q.fail_next_put = True
        try:
            q.put("oops", block=0)
            pytest.fail("The queue didn't fail when it should have")
        except FailingQueueException:
            pass
        _q.fail_next_put = True
        try:
            q.put("oops", timeout=0.1)
            pytest.fail("The queue didn't fail when it should have")
        except FailingQueueException:
            pass
        q.put("last")
        assert qfull(q)
        # Test a failing blocking put
        _q.fail_next_put = True
        try:
            self.do_blocking_test(q.put, ("full",), q.get, ())
            pytest.fail("The queue didn't fail when it should have")
        except FailingQueueException:
            pass
        # Check the Queue isn't damaged.
        # put failed, but get succeeded - re-add
        q.put("last")
        # Test a failing timeout put
        _q.fail_next_put = True
        try:
            self.do_exceptional_blocking_test(
                q.put, ("full", True, 10), q.get, (), FailingQueueException
            )
            pytest.fail("The queue didn't fail when it should have")
        except FailingQueueException:
            pass
        # Check the Queue isn't damaged.
        # put failed, but get succeeded - re-add
        q.put("last")
        assert qfull(q)
        q.get()
        assert not qfull(q)
        q.put("last")
        assert qfull(q)
        # Test a blocking put
        self.do_blocking_test(q.put, ("full",), q.get, ())
        # Empty it
        for i in range(QUEUE_SIZE):
            q.get()
        assert not q.qsize()
        q.put("first")
        _q.fail_next_get = True
        try:
            q.get()
            pytest.fail("The queue didn't fail when it should have")
        except FailingQueueException:
            pass
        assert q.qsize()
        _q.fail_next_get = True
        try:
            q.get(timeout=0.1)
            pytest.fail("The queue didn't fail when it should have")
        except FailingQueueException:
            pass
        assert q.qsize()
        q.get()
        assert not q.qsize()
        _q.fail_next_get = True
        try:
            self.do_exceptional_blocking_test(
                q.get, (), q.put, ("empty",), FailingQueueException
            )
            pytest.fail("The queue didn't fail when it should have")
        except FailingQueueException:
            pass
        # put succeeded, but get failed.
        assert q.qsize()
        q.get()
        assert not q.qsize()

    @pytest.mark.asyncio
    async def test_failing_queue(self):
        # Test to make sure a queue is functioning correctly.
        # Done twice to the same instance.
        q = FailingQueue(QUEUE_SIZE)
        self.failing_queue_test(q)
        self.failing_queue_test(q)
        q.close()
        await q.wait_closed()

    @pytest.mark.asyncio
    async def test_closed_loop_non_failing(self):
        loop = asyncio.get_running_loop()
        _q = janus.Queue(QUEUE_SIZE)
        q = _q.sync_q
        # we are pacthing loop to follow setUp/tearDown agreement
        with patch.object(loop, "call_soon_threadsafe") as func:
            func.side_effect = RuntimeError()
            q.put_nowait(1)
            assert func.call_count == 1
        _q.close()
        await _q.wait_closed()



import asyncio
import contextlib
import threading

import pytest

import janus


class TestMixedMode:
    def test_ctor_noloop(self):
        with pytest.raises(RuntimeError):
            janus.Queue()

    @pytest.mark.asyncio
    async def test_maxsize(self):
        q = janus.Queue(5)
        assert 5 == q.maxsize

    @pytest.mark.asyncio
    async def test_maxsize_named_param(self):
        q = janus.Queue(maxsize=7)
        assert 7 == q.maxsize

    @pytest.mark.asyncio
    async def test_maxsize_default(self):
        q = janus.Queue()
        assert 0 == q.maxsize

    @pytest.mark.asyncio
    async def test_unfinished(self):
        q = janus.Queue()
        assert q.sync_q.unfinished_tasks == 0
        assert q.async_q.unfinished_tasks == 0
        q.sync_q.put(1)
        assert q.sync_q.unfinished_tasks == 1
        assert q.async_q.unfinished_tasks == 1
        q.sync_q.get()
        assert q.sync_q.unfinished_tasks == 1
        assert q.async_q.unfinished_tasks == 1
        q.sync_q.task_done()
        assert q.sync_q.unfinished_tasks == 0
        assert q.async_q.unfinished_tasks == 0
        q.close()
        await q.wait_closed()

    @pytest.mark.asyncio
    async def test_sync_put_async_get(self):
        loop = asyncio.get_running_loop()
        q = janus.Queue()

        def threaded():
            for i in range(5):
                q.sync_q.put(i)

        async def go():
            f = loop.run_in_executor(None, threaded)
            for i in range(5):
                val = await q.async_q.get()
                assert val == i

            assert q.async_q.empty()

            await f

        for i in range(3):
            await go()

        q.close()
        await q.wait_closed()

    @pytest.mark.asyncio
    async def test_sync_put_async_join(self):
        loop = asyncio.get_running_loop()
        q = janus.Queue()

        for i in range(5):
            q.sync_q.put(i)

        async def do_work():
            await asyncio.sleep(1)
            while True:
                await q.async_q.get()
                q.async_q.task_done()

        task = loop.create_task(do_work())

        async def wait_for_empty_queue():
            await q.async_q.join()
            task.cancel()

        await wait_for_empty_queue()

        q.close()
        await q.wait_closed()

    @pytest.mark.asyncio
    async def test_async_put_sync_get(self):
        loop = asyncio.get_running_loop()
        q = janus.Queue()

        def threaded():
            for i in range(5):
                val = q.sync_q.get()
                assert val == i

        async def go():
            f = loop.run_in_executor(None, threaded)
            for i in range(5):
                await q.async_q.put(i)

            await f
            assert q.async_q.empty()

        for i in range(3):
            await go()

        q.close()
        await q.wait_closed()

    @pytest.mark.asyncio
    async def test_sync_join_async_done(self):
        loop = asyncio.get_running_loop()
        q = janus.Queue()

        def threaded():
            for i in range(5):
                q.sync_q.put(i)
            q.sync_q.join()

        async def go():
            f = loop.run_in_executor(None, threaded)
            for i in range(5):
                val = await q.async_q.get()
                assert val == i
                q.async_q.task_done()

            assert q.async_q.empty()

            await f

        for i in range(3):
            await go()

        q.close()
        await q.wait_closed()

    @pytest.mark.asyncio
    async def test_async_join_async_done(self):
        loop = asyncio.get_running_loop()
        q = janus.Queue()

        def threaded():
            for i in range(5):
                val = q.sync_q.get()
                assert val == i
                q.sync_q.task_done()

        async def go():
            f = loop.run_in_executor(None, threaded)
            for i in range(5):
                await q.async_q.put(i)

            await q.async_q.join()

            await f
            assert q.async_q.empty()

        for i in range(3):
            await go()

        q.close()
        await q.wait_closed()

    @pytest.mark.asyncio
    async def test_wait_without_closing(self):
        q = janus.Queue()

        with pytest.raises(RuntimeError):
            await q.wait_closed()

        q.close()
        await q.wait_closed()

    @pytest.mark.asyncio
    async def test_modifying_forbidden_after_closing(self):
        q = janus.Queue()
        q.close()

        with pytest.raises(RuntimeError):
            q.sync_q.put(5)

        with pytest.raises(RuntimeError):
            q.sync_q.get()

        with pytest.raises(RuntimeError):
            q.sync_q.task_done()

        with pytest.raises(RuntimeError):
            await q.async_q.put(5)

        with pytest.raises(RuntimeError):
            q.async_q.put_nowait(5)

        with pytest.raises(RuntimeError):
            q.async_q.get_nowait()

        with pytest.raises(RuntimeError):
            await q.sync_q.task_done()

        await q.wait_closed()

    @pytest.mark.asyncio
    async def test_double_closing(self):
        q = janus.Queue()
        q.close()
        q.close()
        await q.wait_closed()

    @pytest.mark.asyncio
    async def test_closed(self):
        q = janus.Queue()
        assert not q.closed
        assert not q.async_q.closed
        assert not q.sync_q.closed
        q.close()
        assert q.closed
        assert q.async_q.closed
        assert q.sync_q.closed

    @pytest.mark.asyncio
    async def test_async_join_after_closing(self):
        q = janus.Queue()
        q.close()
        with pytest.raises(RuntimeError), contextlib.suppress(asyncio.TimeoutError):
            await asyncio.wait_for(q.async_q.join(), timeout=0.1)

        await q.wait_closed()

    @pytest.mark.asyncio
    async def test_close_after_async_join(self):
        q = janus.Queue()
        q.sync_q.put(1)

        task = asyncio.ensure_future(q.async_q.join())
        await asyncio.sleep(0.1)  # ensure tasks are blocking

        q.close()
        with pytest.raises(RuntimeError), contextlib.suppress(asyncio.TimeoutError):
            await asyncio.wait_for(task, timeout=0.1)

        await q.wait_closed()

    @pytest.mark.asyncio
    async def test_sync_join_after_closing(self):
        q = janus.Queue()
        q.sync_q.put(1)

        q.close()

        loop = asyncio.get_event_loop()
        fut = asyncio.Future()

        def sync_join():
            try:
                q.sync_q.join()
            except Exception as exc:
                loop.call_soon_threadsafe(fut.set_exception, exc)

        thr = threading.Thread(target=sync_join, daemon=True)
        thr.start()

        with pytest.raises(RuntimeError), contextlib.suppress(asyncio.TimeoutError):
            await asyncio.wait_for(fut, timeout=0.1)

        await q.wait_closed()

    @pytest.mark.asyncio
    async def test_close_after_sync_join(self):
        q = janus.Queue()
        q.sync_q.put(1)

        loop = asyncio.get_event_loop()
        fut = asyncio.Future()

        def sync_join():
            try:
                q.sync_q.join()
            except Exception as exc:
                loop.call_soon_threadsafe(fut.set_exception, exc)

        thr = threading.Thread(target=sync_join, daemon=True)
        thr.start()
        thr.join(0.1)  # ensure tasks are blocking

        q.close()

        with pytest.raises(RuntimeError), contextlib.suppress(asyncio.TimeoutError):
            await asyncio.wait_for(fut, timeout=0.1)

        await q.wait_closed()



"""Tests for queues.py"""

import asyncio

import pytest

import janus


class TestQueueBasic:
    async def _test_repr_or_str(self, fn, expect_id):
        """Test Queue's repr or str.

        fn is repr or str. expect_id is True if we expect the Queue's id to
        appear in fn(Queue()).
        """

        _q = janus.Queue()
        q = _q.async_q
        assert fn(q).startswith("<Queue")
        id_is_present = hex(id(q)) in fn(q)
        assert expect_id == id_is_present
        loop = asyncio.get_running_loop()

        async def add_getter():
            _q = janus.Queue()
            q = _q.async_q
            # Start a task that waits to get.
            loop.create_task(q.get())
            # Let it start waiting.
            await asyncio.sleep(0.1)
            assert "_getters[1]" in fn(q)
            # resume q.get coroutine to finish generator
            q.put_nowait(0)

        await add_getter()

        async def add_putter():
            _q = janus.Queue(maxsize=1)
            q = _q.async_q
            q.put_nowait(1)
            # Start a task that waits to put.
            loop.create_task(q.put(2))
            # Let it start waiting.
            await asyncio.sleep(0.1)
            assert "_putters[1]" in fn(q)
            # resume q.put coroutine to finish generator
            q.get_nowait()

        await add_putter()

        _q = janus.Queue()
        q = _q.async_q
        q.put_nowait(1)
        assert "_queue=[1]" in fn(q)

    # def test_repr(self):
    #     self._test_repr_or_str(repr, True)

    # def test_str(self):
    #     self._test_repr_or_str(str, False)

    @pytest.mark.asyncio
    async def test_empty(self):
        _q = janus.Queue()
        q = _q.async_q
        assert q.empty()
        q.put_nowait(1)
        assert not q.empty()
        assert 1 == q.get_nowait()
        assert q.empty()

        assert not _q._sync_mutex.locked()
        _q.close()
        await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_full(self):
        _q = janus.Queue()
        q = _q.async_q
        assert not q.full()

        _q = janus.Queue(maxsize=1)
        q = _q.async_q
        q.put_nowait(1)
        assert q.full()

        assert not _q._sync_mutex.locked()
        _q.close()
        await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_order(self):
        _q = janus.Queue()
        q = _q.async_q
        for i in [1, 3, 2]:
            q.put_nowait(i)

        items = [q.get_nowait() for _ in range(3)]
        assert [1, 3, 2] == items

        assert not _q._sync_mutex.locked()
        _q.close()
        await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_maxsize(self):
        loop = asyncio.get_running_loop()
        _q = janus.Queue(maxsize=2)
        q = _q.async_q
        assert 2 == q.maxsize
        have_been_put = []

        fut = loop.create_future()

        async def putter():
            for i in range(3):
                await q.put(i)
                have_been_put.append(i)
                if i == q.maxsize - 1:
                    fut.set_result(None)
            return True

        async def test():
            t = loop.create_task(putter())
            await fut

            # The putter is blocked after putting two items.
            assert [0, 1] == have_been_put
            assert 0 == q.get_nowait()

            # Let the putter resume and put last item.
            await t
            assert [0, 1, 2] == have_been_put
            assert 1 == q.get_nowait()
            assert 2 == q.get_nowait()

        await test()

        assert not _q._sync_mutex.locked()
        _q.close()
        await _q.wait_closed()


class TestQueueGetTests:
    @pytest.mark.asyncio
    async def test_blocking_get(self):
        _q = janus.Queue()
        q = _q.async_q
        q.put_nowait(1)

        res = await q.get()
        assert 1 == res

        assert not _q._sync_mutex.locked()
        _q.close()
        await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_get_with_putters(self):
        loop = asyncio.get_running_loop()
        _q = janus.Queue(1)
        q = _q.async_q
        q.put_nowait(1)

        fut = loop.create_future()

        async def put():
            t = asyncio.ensure_future(q.put(2))
            await asyncio.sleep(0.01)
            fut.set_result(None)
            return t

        t = await put()

        res = await q.get()
        assert 1 == res

        await t
        assert 1 == q.qsize()

        assert not _q._sync_mutex.locked()
        _q.close()
        await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_blocking_get_wait(self):
        loop = asyncio.get_running_loop()
        _q = janus.Queue()
        q = _q.async_q
        started = asyncio.Event()
        finished = False

        async def queue_get():
            nonlocal finished
            started.set()
            res = await q.get()
            finished = True
            return res

        async def queue_put():
            loop.call_later(0.01, q.put_nowait, 1)
            queue_get_task = loop.create_task(queue_get())
            await started.wait()
            assert not finished
            res = await queue_get_task
            assert finished
            return res

        res = await queue_put()
        assert 1 == res

        assert not _q._sync_mutex.locked()
        _q.close()
        await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_nonblocking_get(self):
        _q = janus.Queue()
        q = _q.async_q
        q.put_nowait(1)
        assert 1 == q.get_nowait()

        _q.close()
        await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_nonblocking_get_exception(self):
        _q = janus.Queue()
        pytest.raises(asyncio.QueueEmpty, _q.async_q.get_nowait)

        assert not _q._sync_mutex.locked()
        _q.close()
        await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_get_cancelled(self):
        loop = asyncio.get_running_loop()
        _q = janus.Queue()
        q = _q.async_q

        async def queue_get():
            return await asyncio.wait_for(q.get(), 0.051)

        async def test():
            get_task = loop.create_task(queue_get())
            await asyncio.sleep(0.01)  # let the task start
            q.put_nowait(1)
            return await get_task

        assert 1 == await test()

        assert not _q._sync_mutex.locked()
        _q.close()
        await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_get_cancelled_race(self):
        loop = asyncio.get_running_loop()
        _q = janus.Queue()
        q = _q.async_q

        f1 = loop.create_future()

        async def g1():
            f1.set_result(None)
            await q.get()

        t1 = loop.create_task(g1())
        t2 = loop.create_task(q.get())

        await f1
        await asyncio.sleep(0.01)
        t1.cancel()

        with pytest.raises(asyncio.CancelledError):
            await t1
        assert t1.done()
        q.put_nowait("a")

        await t2
        assert t2.result() == "a"

        assert not _q._sync_mutex.locked()
        _q.close()
        await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_get_with_waiting_putters(self):
        loop = asyncio.get_running_loop()
        _q = janus.Queue(maxsize=1)
        q = _q.async_q

        loop.create_task(q.put("a"))
        loop.create_task(q.put("b"))

        await asyncio.sleep(0.01)

        assert await q.get() == "a"
        assert await q.get() == "b"

        assert not _q._sync_mutex.locked()
        _q.close()
        await _q.wait_closed()


class TestQueuePut:
    @pytest.mark.asyncio
    async def test_blocking_put(self):
        _q = janus.Queue()
        q = _q.async_q

        # No maxsize, won't block.
        await q.put(1)

        assert not _q._sync_mutex.locked()
        _q.close()
        await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_blocking_put_wait(self):
        loop = asyncio.get_running_loop()
        _q = janus.Queue(maxsize=1)
        q = _q.async_q
        started = asyncio.Event()
        finished = False

        async def queue_put():
            nonlocal finished
            started.set()
            await q.put(1)
            await q.put(2)
            finished = True

        async def queue_get():
            loop.call_later(0.01, q.get_nowait)
            queue_put_task = loop.create_task(queue_put())
            await started.wait()
            assert not finished
            await queue_put_task
            assert finished

        await queue_get()

        assert not _q._sync_mutex.locked()
        _q.close()
        await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_nonblocking_put(self):
        _q = janus.Queue()
        q = _q.async_q
        q.put_nowait(1)
        assert 1 == q.get_nowait()

        assert not _q._sync_mutex.locked()
        _q.close()
        await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_nonblocking_put_exception(self):
        _q = janus.Queue(maxsize=1)
        q = _q.async_q
        q.put_nowait(1)
        pytest.raises(asyncio.QueueFull, q.put_nowait, 2)

        assert not _q._sync_mutex.locked()
        _q.close()
        await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_float_maxsize(self):
        _q = janus.Queue(maxsize=1.3)
        q = _q.async_q
        q.put_nowait(1)
        q.put_nowait(2)
        assert q.full()
        pytest.raises(asyncio.QueueFull, q.put_nowait, 3)

        _q.close()
        await _q.wait_closed()

        _q = janus.Queue(maxsize=1.3)
        q = _q.async_q

        async def queue_put():
            await q.put(1)
            await q.put(2)
            assert q.full()

        await queue_put()

        assert not _q._sync_mutex.locked()
        _q.close()
        await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_put_cancelled(self):
        loop = asyncio.get_running_loop()
        _q = janus.Queue()
        q = _q.async_q

        async def queue_put():
            await q.put(1)
            return True

        async def test():
            return await q.get()

        t = loop.create_task(queue_put())
        assert 1 == await test()
        assert t.done()
        assert t.result()

        assert not _q._sync_mutex.locked()
        _q.close()
        await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_put_cancelled_race(self):
        loop = asyncio.get_running_loop()
        _q = janus.Queue(maxsize=1)
        q = _q.async_q

        put_a = loop.create_task(q.put("a"))
        put_b = loop.create_task(q.put("b"))
        put_c = loop.create_task(q.put("X"))

        await put_a
        assert not put_b.done()

        put_c.cancel()

        with pytest.raises(asyncio.CancelledError):
            await put_c

        async def go():
            a = await q.get()
            assert a == "a"
            b = await q.get()
            assert b == "b"
            assert put_b.done()

        await go()

        assert not _q._sync_mutex.locked()
        _q.close()
        await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_put_with_waiting_getters(self):
        loop = asyncio.get_running_loop()
        fut = loop.create_future()

        async def go():
            fut.set_result(None)
            ret = await q.get()
            return ret

        async def put():
            await q.put("a")

        _q = janus.Queue()
        q = _q.async_q
        t = loop.create_task(go())
        await fut
        await put()
        assert await t == "a"

        assert not _q._sync_mutex.locked()
        _q.close()
        await _q.wait_closed()


class TestLifoQueue:
    @pytest.mark.asyncio
    async def test_order(self):
        _q = janus.LifoQueue()
        q = _q.async_q
        for i in [1, 3, 2]:
            q.put_nowait(i)

        items = [q.get_nowait() for _ in range(3)]
        assert [2, 3, 1] == items

        assert not _q._sync_mutex.locked()
        _q.close()
        await _q.wait_closed()


class TestPriorityQueue:
    @pytest.mark.asyncio
    async def test_order(self):
        _q = janus.PriorityQueue()
        q = _q.async_q
        for i in [1, 3, 2]:
            q.put_nowait(i)

        items = [q.get_nowait() for _ in range(3)]
        assert [1, 2, 3] == items

        assert not _q._sync_mutex.locked()
        _q.close()
        await _q.wait_closed()


class _QueueJoinTestMixin:
    q_class = None

    @pytest.mark.asyncio
    async def test_task_done_underflow(self):
        _q = self.q_class()
        q = _q.async_q
        pytest.raises(ValueError, q.task_done)

        assert not _q._sync_mutex.locked()
        _q.close()
        await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_task_done(self):
        loop = asyncio.get_running_loop()
        _q = self.q_class()
        q = _q.async_q
        for i in range(100):
            q.put_nowait(i)

        accumulator = 0

        # Two workers get items from the queue and call task_done after each.
        # Join the queue and assert all items have been processed.
        running = True

        async def worker():
            nonlocal accumulator

            while running:
                item = await q.get()
                accumulator += item
                q.task_done()

        async def test():
            tasks = [loop.create_task(worker()) for index in range(2)]

            await q.join()
            return tasks

        tasks = await test()
        assert sum(range(100)) == accumulator

        # close running generators
        running = False
        for i in range(len(tasks)):
            q.put_nowait(0)
        await asyncio.wait(tasks)

        assert not _q._sync_mutex.locked()
        _q.close()
        await _q.wait_closed()

    @pytest.mark.asyncio
    async def test_join_empty_queue(self):
        _q = self.q_class()
        q = _q.async_q

        # Test that a queue join()s successfully, and before anything else
        # (done twice for insurance).

        async def join():
            await q.join()
            await q.join()

        await join()

        assert not _q._sync_mutex.locked()
        _q.close()
        await _q.wait_closed()


class TestQueueJoin(_QueueJoinTestMixin):
    q_class = janus.Queue


class TestLifoQueueJoin(_QueueJoinTestMixin):
    q_class = janus.LifoQueue


class TestPriorityQueueJoin(_QueueJoinTestMixin):
    q_class = janus.PriorityQueue
